"""
Making sense of ANSI colors.
"""
import sys
from typing import Literal

try:
    import matplotlib.pyplot as plt
    from matplotlib.ticker import FuncFormatter
except ImportError:
    print("prettypretty.plot requires matplotlib. Please install the package,")
    print("e.g., by executing `pip install matplotlib`, and then")
    print("run `python -m prettypretty.plot` again.")
    sys.exit(1)

import argparse
import math
import pathlib
from typing import Any, cast

from .terminal import Terminal
from .color import close_enough, Color, ColorSpace, ThemeEntry
from .theme import current_sampler, VGA


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="""
            Plot colors on the chroma/hue plane of Oklab while ignoring their
            lightness. If the -i/--input option is specified, this script reads
            newline-separated colors from the named file. If the --vga option is
            specified, it uses the VGA colors. Otherwise, it uses the terminal's
            current ANSI theme colors. This script correctly plots colors
            irrespective of their color space and the current display gamut.
            However, since its visualization library (matplotlib) is limited to
            sRGB, each mark's color is gamut-mapped to sRGB.
        """
    )
    parser.add_argument(
        "-q", "--quiet",
        action="count",
        default=0,
        help="run silently, without printing status updates"
    )
    parser.add_argument(
        "-v", "--verbose",
        action="count",
        default=0,
        help="run in verbose mode, which prints extra information to the console"
    )
    parser.add_argument(
        "--vga",
        action="store_const",
        const=VGA,
        dest="theme",
        help="use VGA colors instead of querying terminal"
    )
    parser.add_argument(
        "-c", "--color",
        action="append",
        dest="colors",
        help="also plot color written as six hexadecimal digits"
    )
    parser.add_argument(
        "-i", "--input",
        help="read newline-separated colors from the named file",
    )
    parser.add_argument(
        "-g", "--gamut",
        action="append",
        dest="gamuts",
        help="also plot boundary of colorspace gamut; value must be sRGB, P3, or Rec2020",
    )
    parser.add_argument(
        "--spectrum",
        action="store_true",
        help="also plot boundary of visible spectrum",
    )
    parser.add_argument(
        "-o", "--output",
        help="write color plot to the named file"
    )
    return parser


class ColorPlotter:
    def __init__(
        self,
        collection_name: None | str = None,
        color_kind: None | str = None,
        gamut_step: None | int = None,
        gamut_range: None | int = None,
        volume: int = 1,
    ) -> None:
        self._collection_name = collection_name
        self._color_kind = color_kind

        # Scattered colors
        self._hues: list[float] = []
        self._chromas: list[float] = []
        self._marks: list[str] = []
        self._mark_colors: list[str] = []

        # Averaged grays as one
        self._grays: list[float] = []
        self._gray_mark = None

        # Lightness bars
        self._lightness: list[float] = []
        self._bar_color: list[str] = []

        # Counts
        self._colors: set[Color] = set()
        self._duplicate_count = 0
        self._base_count = 0
        self._extra_count = 0

        # Gamut boundaries
        self._gamut_step = gamut_step or 1
        self._gamut_range = gamut_range or 20
        self._largest_gamut: None | ColorSpace = None

        self._volume = volume

    def status(self, msg: str) -> None:
        if self._volume >= 1:
            print(msg)

    def detail(self, msg: str) -> None:
        if self._volume >= 2:
            print(msg)

    def start_adding(self) -> None:
        self.status("                Color    L        Chroma   Hue")
        self.status("------------------------------------------------")

    def add(self, name: str, color: Color, marker: str = "o") -> None:
        # Matplotlib is sRGB only
        hex_color = color.to_hex_format()

        # Convert to Oklrch
        oklrch = color.to(ColorSpace.Oklrch)
        lr, c, h = oklrch
        c = round(c, 14)  # Chop off one digit of precision.

        # Update status
        light = f'{lr:.5}'
        if len(light) > 7:
            light = f'{lr:.5f}'
        chroma = f'{c:.5}'
        if len(chroma) > 7:
            chroma = f'{c:.5f}'
        hue = f'{h:.1f}'
        self.status(f"{name:14}  {hex_color}  {light:<7}  {chroma:<7}  {hue:>5}")

        # Handle grays
        if c < 1e-9 or math.isnan(h):
            self._grays.append(lr)
            if self._gray_mark is None:
                self._gray_mark = marker
            elif marker != self._gray_mark:
                raise ValueError(
                    f"inconsistent markers for gray: {marker} vs {self._gray_mark}"
                )
            return

        # Skip duplicates
        if oklrch in self._colors:
            self._duplicate_count += 1
            return

        # Record hue, chroma, color, marker
        h_radian = h * math.pi / 180
        self._hues.append(h_radian)
        self._chromas.append(c)
        self._marks.append(marker)
        self._mark_colors.append(hex_color)

        self._lightness.append(lr)
        self._bar_color.append(Color(ColorSpace.Oklrch, [lr, c, h]).to_hex_format())

        self._colors.add(oklrch)
        if marker == "o":
            self._base_count += 1
        else:
            self._extra_count += 1

    def stop_adding(self) -> None:
        self.status(
            f"\nAltogether {self.format_counts()} colors, "
            f" {len(self._grays)} grays, and {self._duplicate_count} duplicates"
        )

    def format_counts(self) -> str:
        gray_count = len(self._grays)
        if gray_count == 0 and self._extra_count == 0:
            return f"{self._base_count}"

        counts = f"{self._base_count}+{gray_count}"
        if self._extra_count > 0:
            counts += f"+{self._extra_count}"
        return counts

    def effective_max_chroma(self) -> float:
        if self._largest_gamut in (ColorSpace.Srgb, ColorSpace.DisplayP3):
            return 0.4
        elif self._largest_gamut is ColorSpace.Rec2020:
            return 0.5
        elif all(c < 0.3 for c in self._chromas):
            return 0.3
        else:
            return 0.4

    def format_ytick_label_05(self, y: float, _: int) -> str:
        if y % 0.1 < 1e-9 or math.isclose(y, self.effective_max_chroma()):
            return ""
        else:
            return f"{y:.2}"

    def format_ytick_label_10(self, y: float, _: int) -> str:
        if y > 0.01 and y % 0.1 < 1e-9 and y < self.effective_max_chroma() - 0.01:
            return f"{y:.2}"
        else:
            return ""

    def to_point_and_color(self, color: Color) -> tuple[float, float, str]:
        _, c, h = color.to(ColorSpace.Oklrch).coordinates()
        hex_format = Color(ColorSpace.Oklrch, [0.75, c / 3, h]).to_hex_format()

        r, g, b = color.coordinates()
        self.detail(
            f"{color.space()} gamut: "
            f"{r:.2f}, {g:.2f}, {b:.2f} --> "
            f"{c:.5f}, {h:5.1f}º ({hex_format})"
        )

        h = h * math.pi / 180
        return c, h, hex_format

    def generate_boundary_points(
        self, space: ColorSpace, template: list[int], index: int, sign: Literal[1, -1]
    ) -> tuple[list[tuple[float, float]], list[str]]:
        """
        Generate points on the boundary of the given RGB color space.

        This method varies the values of the component with the given index,
        while using the given template's values for the other components. If the
        sign is negative, this method varies the values from (1 - step_size)
        down to 0. Otherwise, it varies the values from (step_size) to 1.
        """
        assert space.is_rgb()

        # Pick the range of steps.
        if sign < 0:
            steps = range(self._gamut_range - self._gamut_step, -1, -self._gamut_step)
        else:
            steps = range(self._gamut_step, self._gamut_range + 1, self._gamut_step)

        points: list[tuple[float, float]] = []
        colors: list[str] = []

        # Iterate over the range.
        for step in steps:
            rgb = [t / self._gamut_range for t in template]
            rgb[index] = step / self._gamut_range

            # Compute the coordinates.
            c, h, hex_format = self.to_point_and_color(Color(space, rgb))
            points.append((c, h))
            colors.append(hex_format)

        return points, colors

    def trace_gamut(
        self, space: ColorSpace
    ) -> tuple[list[tuple[float, float]], list[str]]:
        """
        Trace the boundary of the gamut for the given color space.

        This method traces the boundary by producing coordinates, matplotlib
        path instructions, and gamut-mapped colors for a series of points from
        the red primary to the green primary to the blue primary and back again.
        """
        if (
            self._largest_gamut is None
            or space is ColorSpace.Rec2020
            or space is ColorSpace.DisplayP3 and self._largest_gamut is ColorSpace.Srgb
        ):
            self._largest_gamut = space

        all_points: list[tuple[float, float]]= []
        all_colors: list[str] = []

        def trace(template: list[None | int], index: int, sign: Literal[1, -1]) -> None:
            """Trace primary to secondary or secondary to primary boundary segment."""
            # Scale the template by the number of steps.
            template = [(0 if t is None else t * self._gamut_range) for t in template]

            # Generate the boundary points and add to the complete lists.
            points, colors = self.generate_boundary_points(
                space, cast(list[int], template), index, sign
            )
            all_points.extend(points)
            all_colors.extend(colors)

        hr = "-" * (len(f"{space}") + 54)

        self.detail(hr)
        c, h, hex_format = self.to_point_and_color(Color(space, [1.0, 0.0, 0.0]))
        all_points.append((c, h))
        all_colors.append(hex_format)

        trace([1, None, 0], 1, 1) # red to yellow
        self.detail(hr)
        trace([None, 1, 0], 0, -1) # yellow to green
        self.detail(hr)
        trace([0, 1, None], 2, 1) # green to cyan
        self.detail(hr)
        trace([0, None, 1], 1, -1) # cyan to blue
        self.detail(hr)
        trace([None, 0, 1], 0, 1) # blue to magenta
        self.detail(hr)
        trace([1, 0, None], 2, -1) # magenta to red
        self.detail(hr)

        assert close_enough(all_points[0][0], all_points[-1][0])
        assert close_enough(all_points[0][1], all_points[-1][1])
        return all_points, all_colors

    def add_gamut(self, axes: Any, space: ColorSpace) -> None:
        points, colors = self.trace_gamut(space)

        previous_chroma: None | float = None
        previous_hue: None | float = None
        for (chroma, hue), color in zip(points, colors):
            if previous_chroma is not None:
                assert previous_hue is not None
                axes.plot(  # type: ignore
                    [previous_hue, hue],
                    [previous_chroma, chroma],
                    c = color,
                    lw = 1.5,
                )

            previous_chroma = chroma
            previous_hue = hue

    def add_gamut_label(
        self,
        axes: Any,
        space: ColorSpace,
        label: str,
        green: float,
        blue: float,
        dtext: float,
        pad: int,
    ) -> None:
        _, c, h = Color(space, [0.0, green , blue]).to(ColorSpace.Oklch)
        h = h * math.pi / 180
        axes.annotate(  # type: ignore
            label, xy=(h, c), xytext=(h, c + dtext), textcoords="data",
            bbox=dict(
                pad=pad,
                facecolor="none",
                edgecolor="none"
            ),
            arrowprops=dict(
                arrowstyle="-",
                connectionstyle="arc3",
                shrinkA=0,
                shrinkB=0,
            ),
        )

    def trace_spectrum_boundary(self) -> list[tuple[float, float]]:
        points: list[tuple[float, float]] = []

        # FIXME

        return points

    def create_figure(
        self,
        collection_name: None | str = None,
        color_kind: None | str = None,
        gamuts: None | list[ColorSpace] = None,
        spectrum: bool = False,
    ) -> Any:
        fig: Any = plt.figure(layout="constrained", figsize=(5, 6.5))  # type: ignore
        axes: Any = fig.add_subplot(6, 10, (1, 50), polar=True)
        light_axes: Any = fig.add_subplot(6, 10, (52, 59))

        # Add gamut boundaries if so requested.
        gamuts = gamuts or []
        for space in gamuts:
            self.add_gamut(axes, space)

        # Since markers are shared for all marks in a series, we use a new
        # series for every single color.
        for hue, chroma, color, marker in zip(
            self._hues, self._chromas, self._mark_colors, self._marks
        ):
            size = 80 if marker == "o" else 60
            axes.scatter(
                [hue],
                [chroma],
                c=[color],
                s=[size],
                marker=marker,  # type: ignore
                edgecolors='#000',
                zorder=5,
            )

        if self._grays:
            gray = Color.oklab(sum(self._grays) / len(self._grays), 0.0, 0.0).to_hex_format()

            axes.scatter(
                [0],
                [0],
                c=[gray],
                s=[80],
                marker=self._gray_mark,
                edgecolors='#000',
            )

        axes.set_aspect(1)
        axes.set_rmin(0)
        axes.set_rmax(self.effective_max_chroma())

        # Don't show tick labels at angle
        axes.set_rlabel_position(0)

        # If max chroma is 0.3 or 0.4: matplotlib puts grid circles every 0.05.
        #     To reduce clutter and maximize label utility, place label every
        #     0.10, but start at 0.05. Max chroma at 0.5, matplotlib puts grid
        # If max chroma is 0.5: matplotlib puts grid circles every 0.1. Only
        #     suppress labels for origin and max chroma.
        axes.yaxis.set_major_formatter(FuncFormatter(
                self.format_ytick_label_10 if self.effective_max_chroma() == 0.5
                else self.format_ytick_label_05
        ))

        # Center tick labels on tick
        plt.setp(axes.yaxis.get_majorticklabels(), ha="center")  # type: ignore

        # Make grid appear below points
        axes.set_axisbelow(True)

        # Add label for gamut boundaries
        if ColorSpace.Srgb in gamuts:
            self.add_gamut_label(axes, ColorSpace.Srgb, "sRGB", 1.0, 0.85, 0.12, 0)
        if ColorSpace.DisplayP3 in gamuts:
            self.add_gamut_label(axes, ColorSpace.DisplayP3, "P3", 1.0, 0.95, 0.07, 0)
        if ColorSpace.Rec2020 in gamuts:
            self.add_gamut_label(axes, ColorSpace.Rec2020, "Rec. 2020", 0.95, 1.0, 0.18, 1)

        axes.set_title("Hue & Chroma", style="italic", size=13, x=0.11, y=1.01)

        light_axes.set_yticks([0, 0.5, 1], minor=False)
        light_axes.set_yticks([0.25, 0.75], minor=True)
        light_axes.yaxis.grid(True, which="major")
        light_axes.yaxis.grid(True, which="minor")

        light_axes.bar(
            [x for x in range(len(self._lightness))],
            self._lightness,
            color=self._bar_color,
            zorder=5,
        )
        light_axes.set_ylim(0, 1)
        light_axes.get_xaxis().set_visible(False)

        collection_name = collection_name or self._collection_name
        color_kind = color_kind or self._color_kind or "Colors"
        title = f"{collection_name}: " if collection_name else ""
        title += f"{self.format_counts()} {color_kind} in Oklab"
        fig.suptitle(title, ha="left", x=0.044, weight="bold", size=13)
        fig.text(0.045, 0.085, "Lr", fontdict=dict(style="italic", size=13))

        return fig


def main() -> None:
    options = create_parser().parse_args()
    plotter = ColorPlotter(volume=1-options.quiet+options.verbose)
    terminal_id = None

    gamuts: list[ColorSpace] = []
    for gamut in cast(list[str], options.gamuts or []):
        gamut = gamut.casefold()
        if gamut == 'srgb':
            gamuts.append(ColorSpace.Srgb)
        elif gamut == 'p3':
            gamuts.append(ColorSpace.DisplayP3)
        elif gamut == 'rec2020':
            gamuts.append(ColorSpace.Rec2020)
        else:
            raise ValueError(f"invalid gamut name {gamut}")

    # ----------------------------------------------------------------------------------
    # Prepare Colors for Plotting

    plotter.start_adding()

    if options.input is not None:
        with open(options.input, mode="r", encoding="utf8") as file:
            for color in [Color.parse(l) for l in file.readlines() if l.strip()]:
                plotter.add("", color)
    else:
        with Terminal().cbreak_mode().terminal_theme(options.theme) as term:
            if not options.theme:
                terminal_id = term.request_terminal_identity()

            sampler = current_sampler()
            for index in [0, 1, 3, 2, 6, 4, 5, 7]:
                color = sampler.resolve(index)
                plotter.add(ThemeEntry.try_from_index(index + 2).name(), color)
                color = sampler.resolve(index + 8)
                plotter.add(ThemeEntry.try_from_index(index + 8 + 2).name(), color)

    for color in [Color.parse("#" + c) for c in cast(list[str], options.colors) or []]:
        plotter.add("<extra>", color, marker="d")

    plotter.stop_adding()

    # ----------------------------------------------------------------------------------
    # Labels and file names

    if options.theme:
        collection_name = "VGA"
    elif options.input:
        collection_name = None
    elif terminal_id:
        collection_name = terminal_id[0]
    else:
        collection_name = "Unknown Terminal"

    if options.input:
        color_kind = "Colors"
    else:
        color_kind = "ANSI Colors"

    if options.output is not None:
        file_name = options.output
    elif options.input is not None:
        file_name = pathlib.Path(options.input).with_suffix(".svg")
    else:
        assert collection_name is not None
        file_name = f'{collection_name.replace(" ", "-").lower()}-colors.svg'

    # ----------------------------------------------------------------------------------
    # Create and save plot

    fig = plotter.create_figure(
        collection_name=collection_name,
        color_kind=color_kind,
        gamuts=gamuts,
        spectrum=options.spectrum,
    )
    plotter.status(f"Saving plot to `{file_name}`")
    fig.savefig(file_name, bbox_inches="tight")  # type: ignore


if __name__ == "__main__":
    main()



import argparse
from collections.abc import Iterator
import random
import time

from prettypretty.color import Color
from prettypretty.style import rich, RichText, Style
from prettypretty.terminal import Terminal
from prettypretty.theme import current_sampler


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description='simulate a progress bar'
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--nocolor', action='store_const', const='nocolor', dest='fidelity',
        help='do not use any colors'
    )
    group.add_argument(
        '--ansi', action='store_const', const='ansi', dest='fidelity',
        help='use at most ANSI colors only',
    )
    group.add_argument(
        '--eight-bit', action='store_const', const='eight_bit', dest='fidelity',
        help='use at most 8-bit colors',
    )
    return parser


BLOCKS = ' ▎▌▊█'  # <empty> <partial>+ <full>
STEPS = len(BLOCKS) - 1
WIDTH = 100 // STEPS + (1 if 100 % STEPS != 0 else 0)
assert WIDTH * STEPS >= 100  # Without the adjustment, this wouldn't hold

LIGHT_MODE_BAR = rich().fg(Color.p3(0.0, 1.0, 0.0)).style()
DARK_MODE_BAR = rich().fg(3, 151, 49).style()


def format_bar(percent: float, style: Style) -> RichText:
    """Generate progress bar for given percentage."""
    percent = min(percent, 100)  # Clamp max at 100.0

    # Need integer multiple (full) and index (partial), hence must round
    full, partial = divmod(round(percent), STEPS)
    bar = BLOCKS[-1] * full
    if partial > 0:
        # Only add partial character if it is non-empty
        bar += BLOCKS[partial]
    bar = bar.ljust(WIDTH, BLOCKS[0])

    # Displayed percentage remains nicely floating point
    return RichText.of('  ┫', style, bar, ~style, '┣', f' {percent:5.1f}%')


def progress_reports() -> Iterator[float]:
    percent = 0.0
    while True:
        yield percent
        if percent >= 100.0:
            return
        percent += random.gauss(1, 1/3)


def main() -> None:
    options = create_parser().parse_args()

    with (
        Terminal(fidelity=options.fidelity)
        .terminal_theme()
        .hidden_cursor()
        .scoped_style()
    ) as term:
        style = DARK_MODE_BAR if current_sampler().is_dark_theme() else LIGHT_MODE_BAR
        style = style.prepare(term.fidelity)

        fg = style.foreground
        term.writeln(f'Using {"no color" if fg is None else str(fg)}\n').flush()

        for percent in progress_reports():
            bar = format_bar(percent, style)
            term.column(0).rich_text(bar).flush()
            time.sleep(random.uniform(1/60, 1/10))

        term.writeln('\n').flush()


if __name__ == '__main__':
    main()



from typing import TypeAlias

from .color import (
    AnsiColor,
    DefaultColor,
    EmbeddedRgb,
    GrayGradient,
    TerminalColor,
    TrueColor,
)

IntoTerminalColor: TypeAlias = (
    # The constituent color types
    DefaultColor | AnsiColor | EmbeddedRgb | GrayGradient | TrueColor
    # An 8-bit index
    | int
    # Terminal color itself
    | TerminalColor
)




from collections.abc import Iterator, Sequence
from typing import Self


class DefaultColor:
    """The default foreground or background color."""
    Foreground: DefaultColor = ...
    Background: DefaultColor = ...

    def name(self) -> str: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __le__(self, other: object) -> bool: ...


class AnsiColor:
    """
    The 16 extended ANSI colors.

    The ANSI color comprise eight base colors and their eight bright versions.
    Unlike most other colors, the ANSI colors have no intrinsic color values and
    hence are abstract colors. At the same time, they do have semantics.
    """
    Black: AnsiColor = ...
    Red: AnsiColor = ...
    Green: AnsiColor = ...
    Yellow: AnsiColor = ...
    Blue: AnsiColor = ...
    Magenta: AnsiColor = ...
    Cyan: AnsiColor = ...
    White: AnsiColor = ...
    BrightBlack: AnsiColor = ...
    BrightRed: AnsiColor = ...
    BrightGreen: AnsiColor = ...
    BrightYellow: AnsiColor = ...
    BrightBlue: AnsiColor = ...
    BrightMagenta: AnsiColor = ...
    BrightCyan: AnsiColor = ...
    BrightWhite: AnsiColor = ...

    @staticmethod
    def try_from_8bit(value: int) -> AnsiColor: ...
    def to_8bit(self) -> int: ...
    def is_gray(self) -> bool: ...
    def is_bright(self) -> bool: ...
    def to_3bit(self) -> AnsiColor: ...
    def name(self) -> str: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __le__(self, other: object) -> bool: ...


class EmbeddedRgb:
    """The 6x6x6 RGB cube embedded in 8-bit color."""
    def __new__(cls, r: int, g: int, b:int) -> Self: ...
    @staticmethod
    def try_from_8bit(value: int) -> EmbeddedRgb: ...
    def to_8bit(self) -> int: ...
    def to_24bit(self) -> list[int]: ...
    def to_color(self) -> Color: ...
    def coordinates(self) -> list[int]: ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> int: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...


class GrayGradient:
    """The 24-step gray gradient embedded in 8-bit color."""
    def __new__(cls, value: int) -> Self: ...
    @staticmethod
    def try_from_8bit(value: int) -> GrayGradient: ...
    def to_8bit(self) -> int: ...
    def to_24bit(self) -> list[int]: ...
    def to_color(self) -> Color: ...
    def level(self) -> int: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __le__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...


class TrueColor:
    """24-bit RGB colors."""
    def __new__(cls, r: int, g: int, b: int) -> Self: ...
    @staticmethod
    def from_color(color: Color) -> TrueColor: ...
    def to_color(self) -> Color: ...
    def coordinates(self) -> list[int]: ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> int: ...
    def weighted_euclidian_distance(self, other: TrueColor) -> int: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...


class TerminalColor_Default(TerminalColor):
    """The default foreground and background colors."""
    def __new__(cls, color: DefaultColor) -> Self: ...
    @property
    def color(self) -> DefaultColor: ...


class TerminalColor_Ansi(TerminalColor):
    """The 16 extended ANSI colors."""
    def __new__(cls, color: AnsiColor) -> Self: ...
    @property
    def color(self) -> AnsiColor: ...


class TerminalColor_Rgb6(TerminalColor):
    """The 6x6x6 RGB cube embedded in 8-bit colors."""
    def __new__(cls, color: EmbeddedRgb) -> Self: ...
    @property
    def color(self) -> EmbeddedRgb: ...


class TerminalColor_Gray(TerminalColor):
    """The gray gradient embedded in 8-bit colors."""
    def __new__(cls, color: GrayGradient) -> Self: ...
    @property
    def color(self) -> GrayGradient: ...


class TerminalColor_Rgb256(TerminalColor):
    """24-bit RGB colors."""
    def __new__(cls, color: TrueColor) -> Self: ...
    @property
    def color(self) -> TrueColor: ...


class TerminalColor:
    """Terminal colors."""
    Default = TerminalColor_Default
    Ansi = TerminalColor_Ansi
    Rgb6 = TerminalColor_Rgb6
    Gray = TerminalColor_Gray
    Rgb256 = TerminalColor_Rgb256

    @staticmethod
    def from_8bit(color: int) -> TerminalColor: ...
    @staticmethod
    def from_24bit(r: int, g: int, b: int) -> TerminalColor: ...
    @staticmethod
    def from_color(color: Color) -> TerminalColor: ...
    def try_to_8bit(self) -> int: ...
    def try_to_24bit(self) -> list[int]: ...
    def is_default(self) -> bool: ...
    def sgr_parameters(self, layer: Layer) -> list[int]: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...


class Fidelity:
    """
    Levels of support for ANSI escape codes and colors.
    """
    Plain: Fidelity = ...
    NoColor: Fidelity = ...
    Ansi: Fidelity = ...
    EightBit: Fidelity = ...
    Full: Fidelity = ...

    @staticmethod
    def from_color(color: TerminalColor) -> Fidelity: ...
    @staticmethod
    def from_environment(has_tty: bool) -> Fidelity: ...

    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __le__(self, other: object) -> bool: ...
    def __str__(self) -> str: ...


class Layer:
    """The render layer."""
    Foreground: Layer = ...
    Background: Layer = ...

    def offset(self) -> int: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __str__(self) -> str: ...


class ColorSpace:
    """A color space."""
    Srgb: ColorSpace = ...
    LinearSrgb: ColorSpace = ...
    DisplayP3: ColorSpace = ...
    LinearDisplayP3: ColorSpace = ...
    Rec2020: ColorSpace = ...
    LinearRec2020: ColorSpace = ...
    Oklab: ColorSpace = ...
    Oklch: ColorSpace = ...
    Oklrab: ColorSpace = ...
    Oklrch: ColorSpace = ...
    Xyz: ColorSpace = ...

    def is_rgb(self) -> bool: ...
    def is_polar(self) -> bool: ...
    def is_ok(self) -> bool: ...
    def is_bounded(self) -> bool: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __str__(self) -> str: ...


class HueInterpolation:
    """The rule for deciding how to interpolate hues."""
    Shorter: HueInterpolation = ...
    Longer: HueInterpolation = ...
    Increasing: HueInterpolation = ...
    Decreasing: HueInterpolation = ...

    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...


class OkVersion:
    """The version of the Oklab color space."""
    Original: OkVersion = ...
    Revised: OkVersion = ...

    def cartesian_space(self) -> ColorSpace: ...
    def polar_space(self) -> ColorSpace: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...


class Color:
    """High-resolution colors."""
    def __new__(
        cls, space: ColorSpace, coordinates: Sequence[float]
    ) -> Self:
        """
        Create a new color with the color space and coordinates. The coordinates
        must be a sequence with exactly three elements.
        """
    @staticmethod
    def parse(s: str) -> Color: ...
    @staticmethod
    def srgb(r: float, g: float, b: float) -> Color: ...
    @staticmethod
    def p3(r: float, g: float, b: float) -> Color: ...
    @staticmethod
    def oklab(l: float, a: float, b: float) -> Color: ...
    @staticmethod
    def oklrab(lr: float, a: float, b: float) -> Color: ...
    @staticmethod
    def oklch(l: float, c: float, h: float) -> Color: ...
    @staticmethod
    def oklrch(lr: float, c: float, h: float) -> Color: ...
    @staticmethod
    def from_24bit(r: int, g: int, b: int) -> Color: ...
    def is_default(self) -> bool: ...
    def is_gray(self) -> bool: ...
    def space(self) -> ColorSpace: ...
    def coordinates(self) -> list[float]: ...
    def normalize(self) -> Self: ...
    def to(self, target: ColorSpace) -> Self: ...
    def in_gamut(self) -> bool: ...
    def clip(self) -> Self: ...
    def to_gamut(self) -> Self: ...
    def distance(self, other: Self, version: OkVersion) -> float: ...
    def interpolate(
        self,
        color: Self,
        interpolation_space: ColorSpace,
        interpolation_strategy: HueInterpolation,
    ) -> Interpolator: ...
    def lighten(self, factor: float) -> Self: ...
    def darken(self, factor: float) -> Self: ...
    def contrast_against(self, background: Self) -> float: ...
    def use_black_text(self) -> bool: ...
    def use_black_background(self) -> bool: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> float: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def to_24bit(self) -> list[int]: ...
    def to_hex_format(self) -> str: ...


class Interpolator:
    """A color interpolator."""
    def __new__(
        cls,
        color1: Color,
        color2: Color,
        space: ColorSpace,
        strategy: HueInterpolation
    ) -> Self: ...
    def at(self, fraction: float) -> Color: ...


class ThemeEntry_Default(ThemeEntry):
    """A theme entry for one of the default colors."""
    def __new__(cls, color: DefaultColor) -> Self: ...


class ThemeEntry_Ansi(ThemeEntry):
    """A theme entry for one of the ANSI colors."""
    def __new__(cls, color: AnsiColor) -> Self: ...


class ThemeEntry:
    """The 18 colors in a theme."""
    Default = ThemeEntry_Default
    Ansi = ThemeEntry_Ansi

    @staticmethod
    def try_from_index(index: int) -> ThemeEntry: ...
    def name(self) -> str: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __le__(self, other: object) -> bool: ...


class ThemeEntryIterator:
    """An iterator over theme entries."""
    def __iter__(self) -> Self: ...
    def __next__(self) -> None | ThemeEntry: ...


class Sampler:
    """A color sampler for translating between terminal and high-resolution colors."""
    @staticmethod
    def theme_entries() -> Iterator[ThemeEntry]: ...
    def __new__(cls, version: OkVersion, theme_colors: Sequence[Color]) -> Self: ...
    def __repr__(self) -> str: ...

    """Interrogate color theme."""
    def is_dark_theme(self) -> bool: ...

    """Translate to high-resolution colors."""
    def resolve(
        self,
        color: TerminalColor|DefaultColor|AnsiColor|EmbeddedRgb|GrayGradient|TrueColor|int,
    ) -> Color: ...

    """Translate to ANSI colors."""
    def to_ansi(self, color: Color) -> Color:
        """
        Translate high-resolution to ANSI colors using the best available
        algorithm.
        """
    def supports_hue_lightness(self) -> bool: ...
    def to_ansi_hue_lightness(self, color: Color) -> None | AnsiColor: ...
    def to_closest_ansi(self, color: Color) -> AnsiColor: ...
    def to_ansi_rgb(self, color: Color) -> AnsiColor: ...

    """Translate to 8-bit colors."""
    def to_closest_8bit(self, color: Color) -> TerminalColor: ...
    def to_closest_8bit_with_ansi(self, color: Color) -> TerminalColor: ...

    """Cap terminal colors."""
    def cap(
        self,
        color: TerminalColor|DefaultColor|AnsiColor|EmbeddedRgb|GrayGradient|TrueColor|int,
        fidelity: Fidelity
    ) -> None | TerminalColor: ...


def close_enough(f1: float, f2: float) -> bool: ...



"""
High-level support for terminal styles.

A terminal style covers the full repertoire of text attributes controllable via
ANSI escape codes. In addition to a fluent interface for assembling styles in
the first place, terminal styles support an algebra of inversion and difference
to easily compute the minimal changes necessary for restoring the default
appearance or transitioning to another appearance.

Reifying terminal styles this way not only simplifies updating the terminal, but
it also encourages the definition of terminal styles in a dedicated module. Such
an application-wide style registry greatly simplifies the reuse of styles. They
all are defined in the same module after all. Centralization also helps with
tuning styles so that they harmonize with each other and the design guidelines.
In case where no such guidelines exist, a central module may just help define
them.
"""
from collections.abc import Sequence
import dataclasses
import enum
from typing import cast, Literal, overload, Self, TypeAlias, TypeVar

from .ansi import Ansi
from .color import Color, DefaultColor, Fidelity, Layer, TerminalColor
from .theme import current_sampler


class TextAttribute(enum.Enum):
    """
    The superclass of all enumerations representing what should be orthogonal
    stylistic choices. Each enumeration represents a choice of values, which is
    binary in the common case and ternary for :class:`Weight`. It also encodes
    the default state by aliasing the ``DEFAULT`` attribute. Finally, it encodes
    SGR parameters, which are the values of the enumeration constants.

    :meth:`__invert__` uses the above information to automatically determine
    which text attributes must be set to restore the default appearance of
    terminal output again.
    """
    @property
    def is_default(self) -> bool:
        """Determine whether the text attribute is the default value."""
        return self.value == type(self)['DEFAULT'].value

    def __invert__(self) -> None | Self:
        """
        Invert the text attribute. This method returns the text attribute that
        restores the default appearance again. If this text attribute already
        sets the default state, this method returns ``None``.
        """
        if self.is_default:
            return None
        return type(self)['DEFAULT']


class Weight(TextAttribute):
    """
    The font weight.

    Attributes:
        REGULAR: is the regular, medium weight
        DEFAULT: marks the regular weight as the default
        BOLD: is a heavier weight
        LIGHT: is a lighter weight
    """
    REGULAR = 22
    DEFAULT = 22
    BOLD = 1
    LIGHT = 2


class Slant(TextAttribute):
    """
    The slant.

    Attributes:
        UPRIGHT: is text without slant
        DEFAULT: marks upright text as the default
        ITALIC: is text that is very much slanted
    """
    UPRIGHT = 23
    DEFAULT = 23
    ITALIC = 3


class Underline(TextAttribute):
    """
    Underlined or not.

    Attributes:
        NOT_UNDERLINED: has no inferior line
        DEFAULT: marks the lack of lines as the default
        UNDERLINED: has inferior line
    """
    NOT_UNDERLINED = 24
    DEFAULT = 24
    UNDERLINED = 4


class Overline(TextAttribute):
    """
    Overlined or not.

    Attributes:
        NOT_OVERLINED: has no superior line
        DEFAULT: marks the lack of lines as the default
        OVERLINED: has superior line
    """
    NOT_OVERLINED = 55
    DEFAULT = 55
    OVERLINED = 53


class Strikeline(TextAttribute):
    """
    Stricken or not.

    Attributes:
        NOT_STRICKEN: has no line through
        DEFAULT: marks the lack of lines as the default
        STRICKEN: has a line through
    """
    NOT_STRICKEN = 29
    DEFAULT = 29
    STRICKEN = 9


class Coloring(TextAttribute):
    """
    Reversed or not.

    Attributes:
        NOT_REVERSED: is the inconspicuous, regular order
        DEFAULT: marks the regular order as the default
        REVERSED: reverses foreground and background colors
    """
    NOT_REVERSED = 27
    DEFAULT = 27
    REVERSED = 7


class Visibility(TextAttribute):
    """
    The visibility of text.

    Attributes:
        NOT_HIDDEN: makes text visible
        DEFAULT: marks visible text as the default
        HIDDEN: makes text invisible
    """
    NOT_HIDDEN = 28
    DEFAULT = 28
    HIDDEN = 8


TA = TypeVar('TA', bound=TextAttribute)


def invert_attr(attr: None | TA) -> None | TA:
    """Invert the given text attribute."""
    return None if attr is None else ~attr


def invert_color(color: None | TerminalColor, layer: Layer) -> None | TerminalColor:
    """Invert the given color."""
    if color is None or color.is_default():
        return None
    elif layer is Layer.Foreground:
        return TerminalColor.Default(DefaultColor.Foreground)
    else:
        return TerminalColor.Default(DefaultColor.Background)


class Instruction:
    """
    The superclass of all terminal instructions available to rich text.
    """
    __slots__ = ()

    def delegate(self) -> tuple[str, tuple[object, ...]]:
        raise NotImplementedError()

    @property
    def has_text(self) -> bool:
        return False


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class Style(Instruction):
    """
    A terminal style.

    Attributes:
        weight: for font weight
        slant: for font slant
        underline: for inferior lines
        overline: for superior lines
        strikeline: for lines through
        coloring: for color order
        visibility: for visibility
        foreground: for foreground color
        background: for background color
        fidelity: is the minimum color fidelity and computed automatically,
            with ``None`` indicating unbounded fidelity

    This class captures the state of all text attributes controllable through
    ANSI escape sequences. It distinguishes between:

      * setting an attribute, e.g., when :attr:`underline` is
        :data:`Underline.UNDERLINED`;
      * unsetting an attribute, e.g., when :attr:`underline` is
        :data:`Underline.NOT_UNDERLINED`;
      * ignoring an attribute, e.g., when :attr:`underline` is ``None``.

    Since colors may just have to be downsampled or discarded for rendering even
    if styles only allow color formats directly supported by ANSI escape
    sequences, this class accepts all color formats and spaces while having
    robust facilities for adjusting formats as needed.

    The :attr:`fidelity` attribute is automatically computed during
    initialization and identifies the minimum fidelity level needed for
    rendering this style. A null fidelity indicates that this style contains
    arbitrary colors and hence has unbounded fidelity. Meanwhile
    :attr:`.Fidelity.NOCOLOR` indicates that the style does not contain any
    colors and :attr:`.Fidelity.PLAIN` indicates an empty style.

    .. note::

        Styles overload Python's inversion operator. The result of that
        operation is another style that restores the terminal to its default
        appearance. Styles also overload Python's subtraction operator, which
        returns the style that incrementally transitions from the second to the
        first style. Finally, the string representation of styles is the
        corresponding SGR ANSI escape sequence.

    Instances of this class are immutable.
    """
    weight: None | Weight = None
    slant: None | Slant = None
    underline: None | Underline = None
    overline: None | Overline = None
    strikeline: None | Strikeline = None
    coloring: None | Coloring = None
    visibility: None | Visibility = None
    foreground: None | TerminalColor = None
    background: None | TerminalColor = None
    fidelity: Fidelity = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        if all(
            getattr(self, f.name) is None
            for f in dataclasses.fields(self)
            if f.name != 'fidelity'  # Careful, the attribute has not been set!
        ):
            # No styles
            fidelity = Fidelity.Plain
        elif self.foreground is None and self.background is None:
            # No colors
            fidelity = Fidelity.NoColor
        else:
            # Some color
            fg = self.foreground
            bg = self.background

            # Fill in the fidelity
            fg_fid = Fidelity.Plain if fg is None else Fidelity.from_color(fg)
            bg_fid = Fidelity.Plain if bg is None else Fidelity.from_color(bg)

            fidelity = max(fg_fid, bg_fid)

        object.__setattr__(self, 'fidelity', fidelity)

    @property
    def plain(self) -> bool:
        """The flag for an empty style specification."""
        return self.fidelity is Fidelity.Plain

    def __invert__(self) -> Self:
        if self.plain:
            # Nothing to invert
            return self

        return type(self)(
            weight = invert_attr(self.weight),
            slant = invert_attr(self.slant),
            underline = invert_attr(self.underline),
            overline = invert_attr(self.overline),
            strikeline = invert_attr(self.strikeline),
            coloring = invert_attr(self.coloring),
            visibility = invert_attr(self.visibility),
            foreground = invert_color(self.foreground, Layer.Foreground),
            background = invert_color(self.background, Layer.Background),
        )

    def __sub__(self, other: object) -> Self:
        if not isinstance(other, type(self)):
            return NotImplemented

        inverted_other = ~other
        if inverted_other.plain:
            return self

        return type(self)(
            weight = self.weight or inverted_other.weight,
            slant = self.slant or inverted_other.slant,
            underline = self.underline or inverted_other.underline,
            overline = self.overline or inverted_other.overline,
            strikeline = self.strikeline or inverted_other.strikeline,
            coloring = self.coloring or inverted_other.coloring,
            visibility = self.visibility or inverted_other.visibility,
            foreground = self.foreground or inverted_other.foreground,
            background = self.background or inverted_other.background,
        )

    def __or__(self, other: object) -> Self:
        if not isinstance(other, type(self)):
            return NotImplemented

        return type(self)(
            weight = self.weight or other.weight,
            slant = self.slant or other.slant,
            underline = self.underline or other.underline,
            overline = self.overline or other.overline,
            strikeline = self.strikeline or other.strikeline,
            coloring = self.coloring or other.coloring,
            visibility = self.visibility or other.visibility,
            foreground = self.foreground or other.foreground,
            background = self.background or other.background,
        )

    def prepare(self, fidelity: Fidelity) -> Self:
        """
        Adjust this style specification for rendering with the given fidelity.
        """
        if self.fidelity <= fidelity:
            return self

        if fidelity is Fidelity.Plain:
            # Erase all styles
            return type(self)()

        sampler = current_sampler()

        if self.foreground is None:
            fg = None
        else:
            fg = sampler.cap(self.foreground, fidelity)

        if self.background is None:
            bg = None
        else:
            bg = sampler.cap(self.background, fidelity)

        return dataclasses.replace(self, foreground=fg, background=bg)

    def sgr_parameters(self) -> list[int]:
        """Convert this style to the equivalent SGR parameters."""
        parameters: list[int] = []

        if self.weight is not None:
            parameters.append(self.weight.value)
        if self.slant is not None:
            parameters.append(self.slant.value)
        if self.underline is not None:
            parameters.append(self.underline.value)
        if self.overline is not None:
            parameters.append(self.overline.value)
        if self.strikeline is not None:
            parameters.append(self.strikeline.value)
        if self.coloring is not None:
            parameters.append(self.coloring.value)
        if self.visibility is not None:
            parameters.append(self.visibility.value)
        if self.foreground is not None:
            parameters.extend(self.foreground.sgr_parameters(Layer.Foreground))
        if self.background is not None:
            parameters.extend(self.background.sgr_parameters(Layer.Background))

        return parameters

    def sgr(self) -> str:
        """Convert this style specification into an SGR ANSI escape sequence."""
        return f'{Ansi.CSI}{";".join(str(p) for p in self.sgr_parameters())}m'

    def __str__(self) -> str:
        return self.sgr()

    def delegate(self) -> tuple[str, tuple[object, ...]]:
        return 'write_control', (self.sgr(),)


@dataclasses.dataclass(frozen=True, slots=True)
class Link(Instruction):
    """
    A hyperlink.

    Attributes:
        text:
        href:
        id:
    """
    text: str
    href: str
    id: None | str = None

    def delegate(self) -> tuple[str, tuple[object, ...]]:
        return 'link', (self.text, self.href, self.id)

    @property
    def has_text(self) -> bool:
        return bool(self.text)


@dataclasses.dataclass(frozen=True, slots=True)
class PlaceCursor(Instruction):
    """
    Cursor placement by row and column.

    Attributes:
        row:
        column:
    """
    row: None | int = None
    column: None | int = None

    def delegate(self) -> tuple[str, tuple[object, ...]]:
        return 'at', (self.row, self.column)


@dataclasses.dataclass(frozen=True, slots=True)
class MoveCursor(Instruction):
    """
    Cursor movement along one dimension.

    Attributes:
        move:
        offset:
    """
    move: Literal['up', 'down', 'left', 'right', 'column']
    offset: None | int = None

    def delegate(self) -> tuple[str, tuple[object, ...]]:
        return self.move, (self.offset,)


RichTextElement: TypeAlias = str | Instruction
"""The type of all rich text elements."""


@dataclasses.dataclass(frozen=True, slots=True)
class RichText(Sequence[RichTextElement]):
    """
    The terminal version of rich text mixes text, which is to be rendered
    literally, and style specifications, which are to be rendered as SGR ANSI
    escape codes. The terminal version of rich text also tracks the current
    fidelity level and can be easily prepared for a different fidelity.
    """
    fragments: tuple[RichTextElement, ...]
    fidelity: None | Fidelity = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        fidelity = Fidelity.Plain
        for fragment in self.fragments:
            if isinstance(fragment, str):
                continue
            if isinstance(fragment, Style):
                f = fragment.fidelity
            else:
                f = Fidelity.NoColor

            fidelity = max(f, fidelity)

        object.__setattr__(self, 'fidelity', fidelity)

    @classmethod
    def of(cls, *fragments: RichTextElement) -> Self:
        """Create a rich text object from the the given fragments."""
        return cls(fragments)

    def __getitem__(self, index: int) -> RichTextElement:  # type: ignore
        return self.fragments[index]

    def __len__(self) -> int:
        return len(self.fragments)

    def prepare(self, fidelity: Fidelity) -> Self:
        """Prepare this rich text for rendering at the given fidelity."""
        if self.fidelity is not None and self.fidelity <= fidelity:
            return self

        fragments: list[RichTextElement] = []
        for fragment in self.fragments:
            if isinstance(fragment, str) or fragment.has_text:
                # Keep all text
                pass
            elif fidelity is Fidelity.Plain:
                # Drop instructions on plain fidelity
                continue
            elif isinstance(fragment, Style):
                # Adjust style
                fragment = fragment.prepare(fidelity)
                if fragment.plain:
                    continue

            fragments.append(fragment)

        return type(self)(tuple(fragments))


class rich:
    """
    A rich builder.

    This class helps create styles and rich text through fluent property
    accesses and method invocations. In addition to text attributes and colors,
    it also tracks cursor movements and hyperlinks.

    A rich builder can have isolated or incremental styles. Isolated styles are
    the default for builders created by :meth:`rich`, whereas builders created
    by :meth:`rich.incremental` have incremental styles. The trade-off is that,
    with isolated styles, you don't need to worry about undoing styles because
    this class does that for you. However, you do need to worry about all styles
    standing on their own and cannot rely on the previous style contributing
    attributes. It's just the opposite for incremental styles. You can rely on
    the attributes of the previous style and incrementally add to them. But you
    also need to undo your own styles.

    For isolated and incremental styles alike, if you want to undo the current
    style without setting a new style, you can just call :meth:`undo_style` and
    let prettypretty figure out the needed antistyle. For isolated styles, that
    always is the empty style. For incremental styles, that style may just
    depend on all preceding styles.
    """
    def __init__(self, incremental: bool = False) -> None:
        self._elements: list[None | RichTextElement] = []
        self._last_style: None | int = None
        self._incremental = incremental

    @classmethod
    def incremental(cls) -> Self:
        """Create a new builder with incremental styles."""
        return cls(incremental=True)

    @overload
    def style(self) -> Style:
        ...
    @overload
    def style(self, style: Style) -> Self:
        ...
    def style(self, style: None | Style = None) -> Self | Style:
        """
        Get or add a style.

        If invoked without arguments, this method returns the last updated
        style. It does not consider any preceding styles, even if this rich
        builder is in incremental mode.

        If invoked with a style argument, this method adds the style to the rich
        text sequence being built. Whether the style is treated as isolated or
        incremental depends on this rich builder's mode.
        """
        if style is None:
            if self._last_style is None:
                raise ValueError('rich text without style')
            return cast(Style, self._elements[self._last_style])

        self._elements.append(style)
        self._last_style = len(self._elements) - 1
        return self

    def undo(self) -> Self:
        """Undo the currently effective style."""
        self._elements.append(None)
        return self

    def emit(self, close_style: bool = True) -> RichText:
        """
        Emit the built rich text sequence.

        This method computes effective styles to fill locations that undo the
        current style. As long as the argument allows, it also adds a closing
        style to restore the terminal's default appearance if there isn't one.
        """
        elements: list[RichTextElement] = []
        previous_style = None

        for element in self._elements:
            if element is not None and not isinstance(element, Style):
                elements.append(element)
                continue

            if previous_style is None:
                # Current style is the only style
                if element is not None:
                    elements.append(element)
                    previous_style = element
            elif element is None:
                # Undo previous style, in elements and in previous_style
                elements.append(~previous_style)
                previous_style = None
            elif self._incremental:
                # Incremental mode applies style as is, composes previous_style
                elements.append(element)
                previous_style = element | previous_style
            else:
                # Isolated mode applies style difference, sets previous_style
                elements.append(element - previous_style)
                previous_style = element

        if close_style and previous_style is not None and not previous_style.plain:
            elements.append(~previous_style)

        self._elements = []
        return RichText(tuple(elements))

    def _prepare_style(self) -> Style:
        if not self._elements or not isinstance(self._elements[-1], Style):
            self._elements.append(Style())
            self._last_style = len(self._elements) - 1

        style = self._elements[-1]
        assert isinstance(style, Style)
        return style

    def _handle_text_attribute(self, attribute: TextAttribute) -> Self:
        self._elements[-1] = dataclasses.replace(
            self._prepare_style(),
            **{type(attribute).__name__.lower(): attribute}
        )
        return self

    @property
    def regular(self) -> Self:
        """Update style with regular weight."""
        return self._handle_text_attribute(Weight.REGULAR)

    @property
    def light(self) -> Self:
        """Update style with light weight."""
        return self._handle_text_attribute(Weight.LIGHT)

    @property
    def bold(self) -> Self:
        """Update style with bold weight."""
        return self._handle_text_attribute(Weight.BOLD)

    @property
    def upright(self) -> Self:
        """Update style with upright."""
        return self._handle_text_attribute(Slant.UPRIGHT)

    @property
    def italic(self) -> Self:
        """Update style with italic."""
        return self._handle_text_attribute(Slant.ITALIC)

    @property
    def not_underlined(self) -> Self:
        """Update style with not underlined."""
        return self._handle_text_attribute(Underline.NOT_UNDERLINED)

    @property
    def underlined(self) -> Self:
        """Update style with underlined."""
        return self._handle_text_attribute(Underline.UNDERLINED)

    @property
    def not_overlined(self) -> Self:
        """Update style with not overlined."""
        return self._handle_text_attribute(Overline.NOT_OVERLINED)

    @property
    def overlined(self) -> Self:
        """Update style with overlined."""
        return self._handle_text_attribute(Overline.OVERLINED)

    @property
    def not_stricken(self) -> Self:
        """Update style with not stricken."""
        return self._handle_text_attribute(Strikeline.NOT_STRICKEN)

    @property
    def stricken(self) -> Self:
        """Update style with stricken."""
        return self._handle_text_attribute(Strikeline.STRICKEN)

    @property
    def not_reversed(self) -> Self:
        """Update style with background and foreground colors reversed."""
        return self._handle_text_attribute(Coloring.NOT_REVERSED)

    @property
    def reversed(self) -> Self:
        """Update style with background and foreground colors reversed."""
        return self._handle_text_attribute(Coloring.REVERSED)

    @property
    def not_hidden(self) -> Self:
        """Update style with not hidden."""
        return self._handle_text_attribute(Visibility.NOT_HIDDEN)

    @property
    def hidden(self) -> Self:
        """Update style with hidden."""
        return self._handle_text_attribute(Visibility.HIDDEN)

    @overload
    def fg(self, color: int, /) -> Self:
        ...
    @overload
    def fg(self, c1: int, c2: int, c3: int, /) -> Self:
        ...
    @overload
    def fg(self, color: Color, /) -> Self:
        ...
    @overload
    def fg(self, color: TerminalColor, /) -> Self:
        ...
    def fg(
        self,
        c1: int | Color | TerminalColor,
        c2: None | int = None,
        c3: None | int = None,
    ) -> Self:
        """Update style with foreground color."""
        if isinstance(c1, int):
            if c2 is not None:
                assert c3 is not None
                c1 = TerminalColor.from_24bit(c1, c2, c3)
            else:
                assert c3 is None
                c1 = TerminalColor.from_8bit(c1)
        elif isinstance(c1, Color):
            c1 = TerminalColor.from_color(c1)

        self._elements[-1] = dataclasses.replace(self._prepare_style(), foreground=c1)
        return self

    @overload
    def bg(self, color: int, /) -> Self:
        ...
    @overload
    def bg(self, c1: int, c2: int, c3: int, /) -> Self:
        ...
    @overload
    def bg(self, color: Color, /) -> Self:
        ...
    @overload
    def bg(self, color: TerminalColor, /) -> Self:
        ...
    def bg(
        self,
        c1: int | Color | TerminalColor,
        c2: None | int = None,
        c3: None | int = None,
    ) -> Self:
        """Update style with background color."""
        if isinstance(c1, int):
            if c2 is not None:
                assert c3 is not None
                c1 = TerminalColor.from_24bit(c1, c2, c3)
            else:
                assert c3 is None
                c1 = TerminalColor.from_8bit(c1)
        elif isinstance(c1, Color):
            c1 = TerminalColor.from_color(c1)

        self._elements[-1] = dataclasses.replace(self._prepare_style(), background=c1)
        return self

    def link(self, text: str, href: str, id: None | str = None) -> Self:
        """Add hyperlink."""
        self._elements.append(Link(text, href, id))
        return self

    def up(self, offset: None | int = None) -> Self:
        """Move cursor up."""
        self._elements.append(MoveCursor('up', offset))
        return self

    def down(self, offset: None | int = None) -> Self:
        """"Move cursor down."""
        self._elements.append(MoveCursor('down', offset))
        return self

    def left(self, offset: None | int = None) -> Self:
        """Move cursor left."""
        self._elements.append(MoveCursor('left', offset))
        return self

    def right(self, offset: None | int = None) -> Self:
        """Move cursor right."""
        self._elements.append(MoveCursor('right', offset))
        return self

    def column(self, offset: None | int = None) -> Self:
        """Move cursor to the given column."""
        self._elements.append(MoveCursor('column', offset))
        return self

    def at(self, row: None | int = None, column: None | int = None) -> Self:
        """Move cursor to the given position."""
        self._elements.append(PlaceCursor(row, column))
        return self

    def text(self, text: str = '') -> Self:
        """Add the given text."""
        self._elements.append(text)
        return self
